"""init

Revision ID: 54f8985b0ee5
Revises: 
Create Date: 2019-11-04 14:11:49.734358

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "54f8985b0ee5"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass
